import pytest

from iotanbo_py_utils import file_utils
from iotanbo_py_utils.error import ERR
from iotanbo_py_utils.error import VAL


@pytest.fixture(scope="session")
def existing_dir(tmpdir_factory):
    ed = tmpdir_factory.mktemp("existing_dir")
    print("\n*** Created temporary dir for current session: ", ed)
    return ed


@pytest.fixture(scope="session")
def existing_text_file(existing_dir):
    ef = existing_dir.join("existing_file.txt")
    _, err = file_utils.write_text_file_ne(ef, "test")
    assert not err
    print("\n*** Created temporary text file for current session: ", ef)
    return ef


@pytest.fixture(scope="session")
def existing_text_file_symlink(existing_dir, existing_text_file):
    es = existing_dir + "/existing_text_file_symlink.txt"
    _, err = file_utils.create_symlink_ne(existing_text_file, es)
    assert not err
    print("\n*** Created temporary symlink for current session: ", es)
    return es


@pytest.fixture(scope="session")
def not_existing_dir():
    return "/tmp/this_dir_does_not_exist"


@pytest.fixture(scope="session")
def not_existing_file():
    return "/tmp/this_file_does_not_exist"


def test_file_exists_ne(existing_text_file, not_existing_file):
    assert file_utils.file_exists_ne(existing_text_file)
    assert not file_utils.file_exists_ne(not_existing_file)
    # Sad path:
    # Bad parameter test
    assert not file_utils.file_exists_ne(None)
    assert not file_utils.file_exists_ne({"Dummy": "dict"})


def test_dir_exists_ne(existing_dir, not_existing_dir, existing_text_file):
    assert file_utils.dir_exists_ne(existing_dir)
    assert not file_utils.dir_exists_ne(not_existing_dir)
    # Sad path:
    # File instead of dir
    assert not file_utils.dir_exists_ne(existing_text_file)
    # Bad parameter test
    assert not file_utils.dir_exists_ne(None)
    assert not file_utils.dir_exists_ne({"Dummy": "dict"})


def test_symlink_exists_ne(existing_text_file_symlink):
    assert file_utils.symlink_exists_ne(existing_text_file_symlink)


def test_create_remove_symlink_ne(existing_text_file, existing_dir):
    # Create and remove symlink to file
    text_file_symlink = existing_dir + "/file_symlink.txt"
    _, err = file_utils.create_symlink_ne(existing_text_file, text_file_symlink)
    assert not err
    assert file_utils.symlink_exists_ne(text_file_symlink)
    assert file_utils.remove_symlink_ne(text_file_symlink)
    assert not file_utils.symlink_exists_ne(text_file_symlink)

    # Create and remove symlink to dir
    dir_symlink = existing_dir + "/symlink_to_parent_dir"
    _, err = file_utils.create_symlink_ne(existing_dir, dir_symlink)
    assert not err
    assert file_utils.symlink_exists_ne(dir_symlink)
    assert file_utils.remove_symlink_ne(dir_symlink)
    assert not file_utils.symlink_exists_ne(dir_symlink)


def test_path_base_and_leaf():
    path = "/tmp/base/leaf"
    base, leaf = file_utils.path_base_and_leaf(path)
    assert base == "/tmp/base"
    assert leaf == "leaf"
    # Test with forward slash at the end
    path = "/tmp/base/leaf/"
    base, leaf = file_utils.path_base_and_leaf(path)
    assert base == "/tmp/base"
    assert leaf == "leaf"


def test_read_write_text_file_ne(existing_dir):
    tmpfile = existing_dir + "/tmpfile.txt"
    _, err = file_utils.write_text_file_ne(tmpfile, "test")
    assert not err
    contents, err = file_utils.read_text_file_ne(tmpfile)
    assert not err
    assert contents == "test"
    file_utils.remove_file_ne(tmpfile)


def test_create_path_ne(existing_dir):
    path = existing_dir + "/some/complicated/path"
    _, err = file_utils.create_path_ne(path)
    assert not err
    assert not file_utils.remove_dir_ne(existing_dir + "/some")[ERR]


def test_copy_file_ne(existing_dir, existing_text_file):
    dest = existing_dir + "/existing_text_file_copy.txt"
    _, err = file_utils.copy_file_ne(existing_text_file, dest)
    assert not err
    # Read created copy to verify its contents
    contents, err = file_utils.read_text_file_ne(dest)
    assert contents == "test"
    # Remove created copy
    assert not file_utils.remove_file_ne(dest)[ERR]


def test_move_file_ne(existing_dir):
    # Create temp file
    orig_file = existing_dir + "move_file_test.txt"
    dest = existing_dir + "moved_file.txt"
    contents = "move_file_test.txt"
    _, err = file_utils.write_text_file_ne(orig_file, contents)
    assert not err
    # Move it
    _, err = file_utils.move_file_ne(orig_file, dest)
    assert not err
    # Assert original file not exists
    assert not file_utils.file_exists_ne(orig_file)
    # Assert moved file exists and has correct contents
    assert file_utils.file_exists_ne(dest)
    file_contents, err = file_utils.read_text_file_ne(dest)
    assert contents == file_contents
    # Remove temp file
    assert file_utils.remove_file_ne(dest)


def test_copy_dir_ne(existing_dir):
    # Create path
    orig_path = existing_dir + "/copy/dir/test"
    _, err = file_utils.create_path_ne(orig_path)
    assert not err
    src = existing_dir + "/copy"
    dest = existing_dir + "/dest"
    # Copy tree
    assert not file_utils.copy_dir_ne(src, dest)[ERR]
    # Assert both exist
    assert file_utils.dir_exists_ne(orig_path)
    assert file_utils.dir_exists_ne(dest + "/dir/test")
    # Remove src and dest
    assert not file_utils.remove_dir_ne(src)[ERR]
    assert not file_utils.remove_dir_ne(dest)[ERR]


def test_move_dir_ne(existing_dir):
    # Create path
    orig_path = existing_dir + "/move/dir/test"
    _, err = file_utils.create_path_ne(orig_path)
    assert not err
    src = existing_dir + "/move"
    dest = existing_dir + "/dest"
    # Copy tree
    _, err = file_utils.move_dir_ne(src, dest)
    assert not err
    # Assert only dest exists
    assert not file_utils.dir_exists_ne(orig_path)
    assert file_utils.dir_exists_ne(dest + "/dir/test")
    # Remove dest
    _, err = file_utils.remove_dir_ne(dest)
    assert not err


def test_get_subdirs(existing_dir):
    # Create path
    path1 = existing_dir + "/subdirs/test1/test1_1"
    assert not file_utils.create_path_ne(path1)[ERR]
    path2 = existing_dir + "/subdirs/test2/test1_2"
    assert not file_utils.create_path_ne(path2)[ERR]
    subdirs_list, err = file_utils.get_subdirs_ne(existing_dir + "/subdirs")
    # There must be no error when getting subdirs
    assert not err
    # There must be 2 subdirs, check their names
    assert len(subdirs_list) == 2
    assert "test1" in subdirs_list
    assert "test2" in subdirs_list
    # Cleanup
    assert not file_utils.remove_dir_ne(existing_dir + "/subdirs")[ERR]


def test_get_file_list(existing_dir):
    # Create path
    path = existing_dir + "/file_list_test"
    assert not file_utils.create_path_ne(path)[ERR]
    # Create a file, a dir and a symlink in that dir
    file_utils.create_path_ne(path + "/test_dir")
    file_utils.write_text_file_ne(path + "/test_file.txt", "test_file.txt")
    file_utils.create_symlink_ne(path + "/test_file.txt", path + "/test_file_symlink.txt")
    # Assert that only file is recognized as file
    file_list, err = file_utils.get_file_list_ne(path)
    assert not err
    assert len(file_list) == 2
    assert "test_file.txt" in file_list
    assert "test_file_symlink.txt" in file_list

    # Cleanup
    file_utils.remove_dir_ne(path)


def test_get_total_items(existing_dir):
    """
    Tests:
    get_total_items,
    dir_empty,
    get_item_type
    """
    # Create path
    path = existing_dir + "/get_total_items_test"
    assert not file_utils.create_path_ne(path)[ERR]
    # Create a file, a dir and a symlink in that dir
    file_utils.create_path_ne(path + "/test_dir")
    file_utils.write_text_file_ne(path + "/test_file.txt", "test_file.txt")
    file_utils.create_symlink_ne(path + "/test_file.txt", path + "/test_file_symlink.txt")
    # Assert there are 3 items
    total_items, err = file_utils.get_total_items_ne(path)
    assert not err
    assert total_items == 3
    # test_dir_empty test
    assert not file_utils.dir_empty_ne(path)
    assert file_utils.dir_empty_ne(path + "/not_exists")
    assert file_utils.dir_empty_ne(path + "/test_dir")
    # get_item_type test
    assert file_utils.get_item_type_ne(path)[VAL] == "dir"
    assert file_utils.get_item_type_ne(path + "/test_file.txt")[VAL] == "file"
    assert file_utils.get_item_type_ne(path + "/test_file_symlink.txt")[VAL] == "symlink"
    assert not file_utils.get_item_type_ne(path + "/not_exists")[VAL]
    # Cleanup
    file_utils.remove_dir_ne(path)


def test_env_variables():
    """
    set_env_var,
    get_env_var,
    unset_env_var,
    env_var_exists
    """
    env_var = "dummy_env_var"
    assert not file_utils.env_var_exists(env_var)
    file_utils.set_env_var(env_var, "exists")
    assert file_utils.env_var_exists(env_var)
    assert file_utils.get_env_var(env_var) == "exists"
    file_utils.unset_env_var(env_var)
    assert not file_utils.env_var_exists(env_var)


def test_tar_gz(existing_dir):
    """
    unzip_tar_gz,
    zip_file_tar_gz,
    zip_dir_tar_gz
    """

    # Create a directory with a text file
    path = existing_dir + "/tar_gz_test"
    assert not file_utils.create_path_ne(path)[ERR]
    # Create a dir and a file
    src_dir = path + "/test_dir"
    file_utils.create_path_ne(src_dir + "/inner_dir")
    src_file = src_dir + "/test_file.txt"
    src_file_contents = "test_file.txt"
    file_utils.write_text_file_ne(src_file, src_file_contents)

    # Create .tar.gz archive from file
    assert not file_utils.zip_file_tar_gz(src_file, path + "/result_file.tar.gz")[ERR]

    # Create .tar.gz archive from directory
    assert not file_utils.zip_dir_tar_gz(src_dir, path + "/result_dir.tar.gz")[ERR]

    # Unzip .tar.gz and verify contents
    assert not file_utils.unzip_tar_gz(path + "/result_file.tar.gz",
                                       path + "/unzipped")[ERR]
    assert file_utils.file_exists_ne(path + "/unzipped/test_file.txt")

    file_contents, err = file_utils.read_text_file_ne(path + "/unzipped/test_file.txt")
    assert file_contents == src_file_contents

    # Unzip .tar.gz dir and verify contents
    assert not file_utils.unzip_tar_gz(path + "/result_dir.tar.gz",
                                       path + "/unzipped")[ERR]
    assert file_utils.dir_exists_ne(path + "/unzipped/test_dir")
    assert file_utils.file_exists_ne(path + "/unzipped/test_dir/test_file.txt")
    file_contents, err = file_utils.read_text_file_ne(path + "/unzipped/test_dir/test_file.txt")
    assert file_contents == src_file_contents

    # Cleanup
    assert not file_utils.remove_dir_ne(existing_dir + "/tar_gz_test")[ERR]


def test_get_user_home_dir():
    user_home_dir = file_utils.get_user_home_dir()
    # print(f"\nUser home dir: {user_home_dir}\n")
    assert file_utils.dir_exists_ne(user_home_dir)


# # This test requires internet access, so it is normally disabled
# def test_download_into_file(existing_dir):
#     dest_file = existing_dir + '/tmp.md'
#     header, err = file_utils.download_into_file("https://github.com/iotanbo/
#     cpplibhub/blob/master/docs/GET_STARTED.md",
#                                                 dest_file)
#     assert not err
#     # Cleanup
#     file_utils.remove_file_ne(dest_file)



"""
String utils
"""
import re


class StrUtils:

    # Case conversions
    # https://stackoverflow.com/a/1176023/3824328
    @staticmethod
    def camel_to_snake(name: str) -> str:
        """
        Convert camel-case string into snake-case string.
        :param name: string like 'CamelCaseString'
        :return: string like 'camel_case_string'
        """
        name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', name).lower()

    @staticmethod
    def snake_to_camel(name: str) -> str:
        """
        Convert snake-case string into camel-case.
        :param name: string like 'camel_case_string'
        :return: string like 'CamelCaseString'
        """
        return ''.join(word.title() for word in name.split('_'))

    @staticmethod
    def strip_symbol(text: str, left_side_sym: str, right_side_sym: str,
                     strip_whitespace_before: bool = True,
                     strip_whitespace_after: bool = False) -> str:
        """
        Remove 'left_side_sym' from the beginning and 'right_side_sym' from the end of the text
        if those symbols exist.
        Strip whitespaces according to params.
        Return stripped string.


        :param text:
        :param left_side_sym:
        :param right_side_sym:
        :param strip_whitespace_before:
        :param strip_whitespace_after:
        :return: str
        """
        if strip_whitespace_before:
            text = text.strip()
        text = text.lstrip(left_side_sym)
        text = text.rstrip(right_side_sym)
        if strip_whitespace_after:
            text = text.strip()
        return text



"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -miotanbo_py_utils` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``iotanbo_py_utils.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``iotanbo_py_utils.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""

import iotanbo_py_utils


def main(*_, **__):
    """
    Report version
    """
    print(f"iotanbo_py_utils version {iotanbo_py_utils.__version__}")
    return 0



import ntpath
import os
import shutil
import subprocess  # for execute_shell_cmd
import tarfile
import urllib.request
from pathlib import Path
from urllib import error

from iotanbo_py_utils.error import IotanboError
from iotanbo_py_utils.error import ResultTuple


def file_exists_ne(path_to_file: str) -> bool:
    """
    Check if file exists.
    :return: True if file exists, false if not exists or is a directory.
    :except: None
    """
    try:
        result = os.path.isdir(path_to_file)
        if result:
            return False
        result = os.path.exists(path_to_file)
        return result
    except Exception:
        return False


# def file_exists(path_to_file: str) -> bool:
#     """
#     Check if file exists.
#     :param path_to_file: string
#     :return: True if file exists, false if not exists or is a directory.
#     :except: IotanboError if any kind of exception occurred during check
#     """
#     try:
#         result = os.path.isdir(path_to_file)
#         if result:
#             return False
#         result = os.path.exists(path_to_file)
#         return result
#     except Exception as e:
#         raise IotanboError(str(e)) from e


def dir_exists_ne(path_to_dir: str) -> bool:
    """
    Check if dir exists.
    :param path_to_dir:
    :return: True if dir exists, false if not exists or is a file.
    :except: None
    """
    try:
        return os.path.isdir(path_to_dir)
    except Exception:
        pass
    return False


# def dir_exists(path_to_dir: str) -> bool:
#     """
#     Check if dir exists.
#     :param path_to_dir:
#     :return: True if dir exists, false if not exists or is a file.
#     :except: IotanboError if any kind of exception occurred during check
#     """
#     try:
#         return os.path.isdir(path_to_dir)
#     except Exception as e:
#         raise IotanboError(str(e)) from e


def create_symlink_ne(src: str, dest: str) -> ResultTuple:
    """
    Create a symlink to src with name `dest`;
    In case of error, error message is returned as second element of ResultTuple.
    :param src: path to file, directory or symlink
    :param dest: path to symlink to be created
    :return: ResultTuple: (None, ErrorMsg)
            ErrorMsg: str - empty string if success, error message otherwise
    """
    _, err = get_item_type_ne(src)
    if err:
        return None, err
    try:
        os.symlink(src, dest)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def symlink_exists_ne(path: str) -> bool:
    """
    Check if symlink exists;
    No exceptions.
    """
    try:
        return os.path.islink(path)
    except Exception:
        return False


def remove_symlink_ne(path: str) -> ResultTuple:
    """
    Remove symlink from the file system (do not raise exceptions).
    :param path: string
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: str - empty string if the link
                                   was removed or not exists,
                                   or error message otherwise;
    """
    if symlink_exists_ne(path):
        try:
            os.unlink(path)
        except Exception as e:
            return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def path_base_and_leaf(path: str) -> (str, str):
    """
    Split path to a base part and a file or directory name, like in the following example:
    path: '/a/b'; base: '/a'; leaf: 'b'. Return (base: str, leaf: str).
    :raise: IotanboError
    """
    try:
        head, tail = ntpath.split(path)
        if not tail:  # in case there is trailing slash at the end of path
            return ntpath.split(head)[0], ntpath.basename(head)
        return head, tail
    except Exception as e:
        raise IotanboError(str(e)) from e


def path_base_and_leaf_ne(path: str) -> ResultTuple:
    """
    Split path to a base part and a file or directory name.
    Return ResultTuple ((base: str, leaf: str), ErrorMsg)
    """
    try:
        head, tail = ntpath.split(path)
        if not tail:  # in case there is trailing slash at the end of path
            return (ntpath.split(head)[0], ntpath.basename(head)), ""
        return (head, tail), ""
    except Exception as e:
        return None, f"{e.__class__.__name__: {str(e)}}"


def write_text_file_ne(filename: str, contents: str = '', encoding: str = "utf-8") -> ResultTuple:
    """
    Create a new text file and write contents into it (do not raise exceptions).
    If file with specified name already exists, it will be overwritten.
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if success, or error message otherwise.
    """
    try:
        with open(filename, 'w', encoding=encoding) as f:
            f.write(contents)
        return None, ""
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"


def read_text_file_ne(filename: str, encoding: str = "utf-8",) -> ResultTuple:
    """
    Read a text file (do not raise exceptions).
    :param filename: path to file
    :param encoding: text file encoding ("utf-8" default)

    :return: ResultTuple: (Result, ErrorMsg):
                         Result: str - file contents if success;
                         ErrorMsg: str - empty string if success, or error message otherwise.
    """
    result = ""
    if not file_exists_ne(filename):
        return result, "file_not_exists"
    try:
        with open(file=filename, mode='r', encoding=encoding) as f:
            return f.read(), ""
    except Exception as e:
        return result, f"{e.__class__.__name__}: {str(e)}"


def read_text_file_lines_ne(filename: str, encoding: str = "utf-8",
                            line_sep: str = "\n") -> ResultTuple:
    """
    Read a text file (do not raise exceptions).

    :param filename: path to file
    :param encoding: text file encoding ("utf-8" default)
    :param line_sep: line separator char(s)

    :return: ResultTuple: (Result, ErrorMsg):
                         Result: List[str] - list of lines if success;
                         ErrorMsg: str - empty string if success, or error message otherwise.
    """
    result = []
    contents, err = read_text_file_ne(filename, encoding=encoding)
    if err:
        return result, []
    result = contents.split(sep=line_sep)
    return result, ""


def create_path_ne(path: str, overwrite=False) -> ResultTuple:
    """
    Create path in the filesystem (do not raise exceptions).
    :param path: path to be created
    :param overwrite: if true, existing old directory will be overwritten
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: str - empty string if success, or error message otherwise.
    """
    try:
        if dir_exists_ne(path):
            if not overwrite:
                return None, ""
            shutil.rmtree(path)
        os.makedirs(path)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def remove_file(path: str) -> None:
    os.remove(path)


def remove_file_ne(path: str) -> ResultTuple:
    """
    Remove the file from the file system (do not raise exceptions).
    :param path:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if the file successfully removed
                                   and/or not exists, or error message otherwise;
    """
    if file_exists_ne(path):
        try:
            os.remove(path)
        except Exception as e:
            return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def remove_dir_ne(path) -> ResultTuple:
    """
    Remove directory and all its contents (a tree) from the file system (do not raise exceptions).
    :param path:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if the directory was removed or not exists,
                       or error message otherwise;
    """
    if dir_exists_ne(path):
        try:
            shutil.rmtree(path)
        except Exception as e:
            return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def copy_file_ne(origin, dest) -> ResultTuple:
    """
    Copy a file (do not raise exception).
    :param origin:
    :param dest:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if existing file was copied,
                         or error message otherwise;
    """
    if not file_exists_ne(origin):
        return None, 'origin does not exist'
    if remove_file_ne(dest)[1]:
        return None, "old file can't be removed"
    try:
        shutil.copy(origin, dest)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def move_file_ne(origin: str, dest: str) -> ResultTuple:
    """
    Move a file (do not raise exception).
    :param origin:
    :param dest:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if existing file was moved,
                         or error message otherwise;
    """
    if not file_exists_ne(origin):
        return None, 'origin does not exist'
    if remove_file_ne(dest)[1]:
        return None, "old file can't be removed"
    try:
        shutil.move(origin, dest)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def copy_dir_ne(origin: str, dest: str) -> ResultTuple:
    """
    Copy a directory and its contents (tree) to the dest, do not raise exception.
    If old 'dest' exists, it will be removed first.
    :param origin:
    :param dest:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if directory was copied
                         and exists or error message otherwise;
    """

    if not dir_exists_ne(origin):
        return None, 'origin does not exist'

    if dir_exists_ne(dest):
        _, err = remove_dir_ne(dest)
        if err:
            return None, f"(file_utils::copy_dir_ne) old directory '{dest}' can't be removed: {err}"
    try:
        shutil.copytree(origin, dest)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def move_dir_ne(origin: str, dest: str) -> ResultTuple:
    """
    Move a directory and its contents (a tree) into a dest (do not raise exception).
    If old 'dest' exists, it will be removed first.
    This function is equivalent to 'rename'.
    :param origin:
    :param dest:
    :return: ResultTuple: (None, ErrorMsg):
                         ErrorMsg: empty string if success or error message otherwise;
    """
    if not dir_exists_ne(origin):
        return None, 'origin does not exist'
    if remove_dir_ne(dest)[1]:
        return None, "old directory can't be removed"
    try:
        shutil.move(origin, dest)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return None, ""


def get_subdirs_ne(path: str) -> ResultTuple:
    """
    Get list of all first child subdirectories that the directory contains.
    Does not raise exceptions.
    :param path:
    :return: ResultTuple: (subdirs, ErrorMsg):
                         subdirs: list of subdirectory names
                         ErrorMsg: empty string if success or error message otherwise;
    """
    if not dir_exists_ne(path):
        return [], 'path not exists'
    # https://stackoverflow.com/questions/973473/getting-a-list-of-all-subdirectories-in-the-current-directory
    try:
        subdirs = [subdir.name for subdir in os.scandir(path) if subdir.is_dir()]
    except Exception as e:
        return [], f"{e.__class__.__name__}: {str(e)}"
    return subdirs, ""


def get_file_list_ne(path: str) -> ResultTuple:
    """
    List of files located in the directory. The list includes symlinks,
    but does not include directories.
    Does not raise exceptions.
    :param path:
    :return: ResultTuple: (file_list, ErrorMsg):
                         file_list: list of files located in the directory;
                         ErrorMsg: empty string if success or error message otherwise;
    """
    if not dir_exists_ne(path):
        return [], 'path not exists'
    try:
        file_list = [file.name for file in os.scandir(path) if file.is_file()]
    except Exception as e:
        return [], f"{e.__class__.__name__}: {str(e)}"
    return file_list, ""


def get_total_items_ne(path: str) -> ResultTuple:
    """
    Total number of child elements in the directory.
    :param path:
    :return: ResultTuple: (total_items: int, ErrorMsg):
                         total_items: number of child items in the directory;
                         ErrorMsg: empty string if success or error message otherwise;
    """
    if not dir_exists_ne(path):
        return 0, 'path not exists'
    try:
        total_items = len([item.name for item in os.scandir(path)])
    except Exception as e:
        return 0, f"{e.__class__.__name__}: {str(e)}"
    return total_items, ""


def dir_empty_ne(path: str) -> bool:
    """
    Check if directory is empty. Do not raise exceptions.
    :param path:
    :return: True if dir is empty or does not exist, False if dir exists and not empty
    """
    try:
        if os.path.exists(path) and os.path.isdir(path):
            if not os.listdir(path):
                return True
            else:
                return False
    except Exception:
        pass
    return True


def get_item_type_ne(path: str) -> ResultTuple:
    """
    Get type of the file system item ('file', 'dir, 'symlink').
    Does not raise exceptions.
    :param path:
    :return: ResultTuple: (Result: str, ErrorMsg):
             Result:  str - the file system item type ('file', 'dir, 'symlink');
                            empty string is returned if type is unknown or item
                            does not exist;
             ErrorMsg: str - empty string if success, error message otherwise
    """
    try:
        if symlink_exists_ne(path):  # symlink must be first because symlink is also a file
            item_type = 'symlink'
        elif file_exists_ne(path):
            item_type = 'file'
        elif dir_exists_ne(path):
            item_type = 'dir'
        else:
            return "", "not exists"
    except Exception as e:
        return "", f"{e.__class__.__name__}: {str(e)}"

    return item_type, ""


def get_user_home_dir() -> str:
    return str(Path.home())


def get_cwd() -> str:
    return os.getcwd()


def get_tree_dir_list(root_dir: str, exclude_dir_list) -> ResultTuple:
    """
    Walk thru dirs inside the 'root_dir' and make a list of all dirs and their sub dirs.
    If a dir path starts with one of the dir paths in the 'exclude_dir_list', it and its contents
    will not be included into the result.
    All paths except 'root_dir' are relative to the 'root_dir'.
    Return list of all sub dirs inside 'root_dir', the paths are relative to the 'root_dir'
    and do not start with dot and slash (./)


    :param root_dir: absolute path to the root dir of the tree
    :param exclude_dir_list: tuple, list or set of paths relative to the 'root_dir',
                             without starting dot and slash (./)
    :return: ResultTuple (List[str], ErrorMsg)
    """
    base_path = root_dir
    base_path_len = len(base_path)
    result = []
    try:
        for path, dirs, files in os.walk(base_path):
            relative_path = path[base_path_len + 1:]
            if not relative_path:  # do not add empty paths
                continue

            excluded = False

            # Check if path is in the exclude_dir_list
            for excluded_path in exclude_dir_list:
                if not excluded_path:  # ignore empty strings
                    continue
                if relative_path.startswith(excluded_path):
                    excluded = True
                    break
            if not excluded:
                # Append path without base_path part to the list
                result.append(relative_path)

    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}"

    return result, ""


def get_tree_file_list(root_dir: str, tree_dir_list: list, excluded_file_list) -> ResultTuple:
    """
    Create a tree file list based on its tree_dir_list;
    File paths starting with any path in 'exclude_file_list' will not be included
    into the result.
    :param root_dir: Absolute path to the tree root dir
    :param tree_dir_list: List of relative paths
    :param excluded_file_list: Tuple, list or set of relative paths
    :return: ResultTuple (list[str], ErrorMsg)
    """

    base_path = root_dir
    result = []
    # Append empty string to tree_dir_list to get also files in the 'root_dir'
    tree_dir_list.append("")
    try:
        for d in tree_dir_list:
            current_dir = base_path + os.sep + d
            files, err = get_file_list_ne(current_dir)
            if err:
                return result, err

            for f in files:
                relative_file_name = d + os.sep + f
                if not d:
                    relative_file_name = f

                # Check if file name is in the exclude_file_list
                excluded = False
                for excluded_file in excluded_file_list:
                    if not excluded_file:
                        continue
                    if relative_file_name == excluded_file:
                        excluded = True
                        break

                if not excluded:
                    result.append(relative_file_name)

    except Exception as e:
        return result, f"{e.__class__.__name__}: {str(e)}"
    tree_dir_list.remove("")

    return result, ""


def get_file_size_ne(path: str) -> ResultTuple:
    """
    Get file size without exceptions.
    :param path: absolute path to the file
    :return: ResultTuple (int, ErrorMsg)
    """
    fsize = 0
    try:
        fsize = os.path.getsize(path)
    except Exception as e:
        return fsize, f"{e.__class__.__name__}: {str(e)}"
    return fsize, ""

# ---------------------------------------------------------------------------------------------------------------------
# Environment variables


def set_env_var(name: str, value: str) -> None:
    os.environ[name] = value


def get_env_var(name: str) -> str:
    if name not in os.environ:
        return ''
    return os.environ[name]


def unset_env_var(name: str) -> None:
    if name in os.environ:
        del os.environ[name]


def env_var_exists(name: str) -> bool:
    return name in os.environ


# ---------------------------------------------------------------------------------------------------------------------
# Execute shell commands

# 'cmd_and_args' must be a list, each argument must be a separate list element
def execute_shell_cmd(cmd_and_args: list):
    subprocess.check_call(cmd_and_args, env=dict(os.environ))


# ---------------------------------------------------------------------------------------------------------------------
# Zip, tar, gzip archives

def unzip_tar_gz(file: str, dest_dir: str, remove_after_extract=False) -> ResultTuple:
    """
    Unzip .tar.gz archive into the `dest_dir`
    :param file: source .tar.gz archive
    :param dest_dir: destination dir
    :param remove_after_extract: bool
    :return: ResultTuple: (None, ErrorMsg):
                 ErrorMsg: empty string if success or error message otherwise;
    """
    if file_exists_ne(dest_dir):
        return None, "destination is not directory: " + str(dest_dir)

    if not file_exists_ne(file):
        return None, "source file not exists"

    if not dir_exists_ne(dest_dir):
        if create_path_ne(dest_dir)[1]:
            return None, "can't create destination directory: " + str(dest_dir)
    try:
        with tarfile.open(file, 'r:*') as t:
            t.extractall(dest_dir)
        if remove_after_extract:
            remove_file_ne(file)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)}, destination dir: {str(dest_dir)}"
    return None, ""


def zip_file_tar_gz(file, dest_file, remove_after=False) -> ResultTuple:
    """
    Create .tar.zip archive from file.
    :param file: source file
    :param dest_file: destination file to be written
    :param remove_after: if True, source file will be removed after archive created
    :return: ResultTuple: (None, ErrorMsg):
                 ErrorMsg: empty string if success or error message otherwise;
    """
    if not file_exists_ne(file):
        return None, "source file not exists: " + str(file)
    try:
        with tarfile.open(dest_file, "w:gz") as tar:
            tar.add(file, arcname=os.path.basename(file))
        if remove_after:
            remove_file(file)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)} when creating file '{str(dest_file)}'"
    return None, ""


def zip_dir_tar_gz(source_dir: str, dest_file: str, remove_after=False) -> ResultTuple:
    """
    Create .tar.zip archive from directory tree.
    :param source_dir:
    :param dest_file:
    :param remove_after: if True, source dir will be removed after archive created
    :return: ResultTuple: (None, ErrorMsg):
                 ErrorMsg: empty string if success or error message otherwise;
    """
    if not dir_exists_ne(source_dir):
        return None, "source directory not exists: " + str(source_dir)
    try:
        with tarfile.open(dest_file, "w:gz") as tar:
            tar.add(source_dir, arcname=os.path.basename(source_dir))
        if remove_after:
            remove_dir_ne(source_dir)
    except Exception as e:
        return None, f"{e.__class__.__name__}: {str(e)} when creating file '{str(dest_file)}'"
    return None, ""


# ---------------------------------------------------------------------------------------------------------------------
# File download

def download_into_file(url, dest_file) -> ResultTuple:
    """
    Download url synchronously and save result into file.
    :param url:
    :param dest_file: file to write url contents
    :return: Dict:
            ["response_header"]: str - http response header if success
            ["error"]: str - empty string if success or error message otherwise
    :return: ResultTuple: (response_header: str, ErrorMsg):
                 response_header: str - http response header if success
                 ErrorMsg: empty string if success or error message otherwise;
    """
    response_header = ""
    try:
        response_header = urllib.request.urlretrieve(url, dest_file)[1]
    except error.URLError as e:
        return None, f"{e.__class__.__name__}: {str(e)}"
    return response_header, ""


# ---------------------------------------------------------------------------------------------------------------------
# Git operations (simplified)

def git_clone(git_url: str, path: str) -> None:
    if dir_exists_ne(path):
        remove_dir_ne(path)
    cmd = ['git', 'clone', git_url, path]
    execute_shell_cmd(cmd)


def git_pull(path) -> None:
    if not dir_exists_ne(path):
        return
    old_path = os.getcwd()
    os.chdir(path)
    cmd = ['git', 'pull']
    execute_shell_cmd(cmd)
    os.chdir(old_path)



"""
Custom exceptions for iotanbo_py_utils
"""

from typing import Any

# Error message type alias
ErrorMsg = str

"""
ResultTuple allows returning both result and error message from a function without raising exceptions.
How it works:
* The result of function execution is returned as the first element of the tuple.
* The error message is returned as the second element of the tuple.
* If ErrorMsg is an empty string, then there is no error.
"""
ResultTuple = (Any, ErrorMsg)

# Indexes of elements inside the ResultTuple
VAL = 0  # value is the first element
ERR = 1  # error message is the second element


class IotanboError(Exception):
    """
    Generic error with an optional message, base class for
    other package-specific errors
    """
    pass
